function [price] = pi_fft_asian_gbm(s0,strikes,sigma,r,q,T,dt,N)

%% Parameters
mu = r-q; % Arbitrage-free drift of S_t
m = round(T/dt); % Number of time steps
T = m*dt; % Avoiding rounding error
% Mean and volatility of z_1 and transitions
mu_z1 = (mu-0.5*sigma^2)*dt;
sigma_z1 = sigma*sqrt(dt);

%% Initialising grid
% Defining the z range based on chebyshev's inequality
k = 10; % 1-1/k^2 of the probability included. 99 %
zmin = T*(mu-0.5*sigma^2) - k*sqrt(T*sigma^2);
zmax = T*(mu-0.5*sigma^2) + k*sqrt(T*sigma^2);
% Initialising grid
dz = (zmax-zmin)/(N-1);
z = (zmin:dz:zmax);
ymax = 1.2651e+04 + 4000;
ymin = 1.2651e+04 - 4000;
dy = (ymax-ymin)/(N-1);
y = ymin:dy:ymax;
% Defining the y,z grid
[Y,Z] = meshgrid(y,z);

%% Defining the transformation variable, z_1 = u(y_2|z_2), for f_{A_2|Z_2}
u = 3*Y/s0 - 1 - exp(Z); % u(y_2|z_2) = ln(tranz1)
correction = (u <= 0); % because e^z_1 > 0
u(correction) = 10^(-60);
u = log(u); % transz1 = u(y_2|z_2)
ind = ones(N);
ind(correction) = 0;
uder = 3./(3*Y-s0-s0*exp(Z));

%% Calculating the inital desity f_{A_2,Z_2}(y_2,z_2):
% f_{Z1}(z2-z1):
f_trans = (1/(sqrt(2*pi)*sigma_z1)) * exp(-0.5*(((Z-u)-mu_z1).^2)/...
    (sigma_z1^2)).*ind;
% f_{Z_1}((u(y_2|z_2)):
f = (1/(sqrt(2*pi)*sigma_z1)) * exp(-0.5*((u-mu_z1).^2)/(sigma_z1^2)).*ind;
% f_{A_2,Z_2}(y_2,z_2) (overwrites f_{Z_1} which is no longer needed)
f = f_trans.*f.*abs(uder);
neg = (f < 0);
f(neg) = 0;
% Scaling - Securing that the joint pdf integrates to 1
w = trapz(z,f,1);
w = trapz(y,w);
f = f/w;

%% Initialising the characteristic function of the transition prob.
freq = mod(1/2+(0:(N-1))/(N),1)-1/2;
v = freq * (2*pi/dz);
v = v';
phi = exp(-0.5*sigma_z1^2*v.^2).*exp(-1i*v*mu_z1);
phi_m = repmat(phi,1,length(phi));

%% Recursions
for j = 3:m
    % Interpolating, setting f_i-1
    YI = (j/(j-1))*Y - s0/(j-1)*exp(Z);
    f = interp2(Y,Z,f,YI,Z,'spline');
    corr = isnan(f);
    f(corr) = 0;
    
    % Calculating the convolution
    f = fft(f);
    f = (phi_m.*f);
    f = (real(ifft(f)));
    neg = (f<=0);
    f(neg) = 0;

% Making sure the final joint density integrates to 1
    w = trapz(y,f,1);
    w = trapz(z,w);
    f = f/w;
    
%     if (mod(j,5) == 0)
%     figure;
%     surf(y,z,f)
%     end
end

%% Calculating the risk neutral expectation, using the trapesoidal rule
V = trapz(z,f);
price = zeros(1,length(strikes));
for i = 1:length(strikes)
    func = max(y - strikes(i),0).*V;
    price(1,i) = exp(-r*T)*trapz(y,func);
end

end