function [price] = pi_fft_asian_nig(s0,strikes,alpha,delta,beta,mu,r,q,T,dt,N)

%% Parameters
omega = delta*( sqrt(alpha^2-(beta+1)^2) - sqrt(alpha^2-beta^2) );
mu = (r - q + omega); % Mean correction
m = round(T/dt); % Number of time steps
T = m*dt; % To avoid rounding error

%% Initialising grid
% Defining the z range based on chebyshev's inequality (approx for NIG,
% based on GBM)
k = 10; % 1-1/k^2 of the probability included. 99 %
zmin = T*(mu + delta*beta/(sqrt(alpha^2-beta^2))) - k*...
    alpha*sqrt(delta*T*(alpha^2-beta^2)^(-3/2) );
zmax = T*(mu + delta*beta/(sqrt(alpha^2-beta^2))) + k*...
    alpha*sqrt(delta*T*(alpha^2-beta^2)^(-3/2) );
% Initialising grid
dz = (zmax-zmin)/(N-1);
z = (zmin:dz:zmax);
ymax = 16651; % Approx. mean +/- 4000 (Set from observations of MC sims)
ymin = 8651;
dy = (ymax-ymin)/(N-1);
y = ymin:dy:ymax;
% Defining the y,z grid
[Y,Z] = meshgrid(y,z);

%% Defining the transformation variable, z_1 = u(y_2|z_2), for f_{A_2|Z_2}
u = 3*Y/s0 - 1 - exp(Z); % u(y_2|z_2) = ln(tranz1)
correction = (u <= 0); % because e^z_1 > 0
u(correction) = 10^(-60);
u = log(u); % transz1 = u(y_2|z_2)
ind = ones(N);
ind(correction) = 0;
uder = 3./(3*Y-s0-s0*exp(Z));

%% Calculating the inital desity f_{A_2,Z_2}(y_2,z_2):

% f_{Z1}(z2-z1):
f_trans = alpha*dt*delta/pi * exp( dt*delta*sqrt(alpha^2-beta^2) +...
    beta*(Z-u-dt*mu) ) .* besselk(1,alpha*sqrt(dt^2*delta^2 +...
    (Z-u-dt*mu).^2) ) ./ (sqrt(dt^2*delta^2 + (Z-u-dt*mu).^2)).*ind;
% f_{Z_1}((u(y_2|z_2)):
f = alpha*dt*delta/pi * exp( dt*delta*sqrt(alpha^2-beta^2) +...
    beta*(u-dt*mu) ) .* besselk(1,alpha*sqrt(dt^2*delta^2 +...
    (u-dt*mu).^2) ) ./ (sqrt(dt^2*delta^2 + (u-dt*mu).^2)).*ind;
% f_{A_2,Z_2}(y_2,z_2) (overwrites f_{Z_1} which is no longer needed)
f = f_trans.*f.*abs(uder);
neg = (f < 0); % Indecies to negative values
f(neg) = 0;
nan = isnan(f);
f(nan) = 0;
% Scaling - Securing that the joint pdf integrates to 1
w = trapz(z,f,1);
w = trapz(y,w);
f = f/w;

%% Initialising the characteristic function of the transition prob.
freq = mod(1/2+(0:(N-1))/(N),1) - 1/2;
v = freq * (2*pi/dz);
v = v';
phi = conj(exp( -delta*dt*( sqrt(alpha^2-(beta+1i*v).^2 ) -...
    sqrt(alpha^2-beta^2) ) + 1i*v*mu*dt ));
phi_m = repmat(phi,1,length(phi));

%% Recursions
for j = 3:m
    % Interpolating, setting f_i-1
    YI = (j/(j-1))*Y - s0/(j-1)*exp(Z);
    f = interp2(Y,Z,f,YI,Z,'spline');
    corr = isnan(f);
    f(corr) = 0;
    
    % Calculating the convolution
    f = fft(f);
    f = (phi_m.*f);
    f = (real(ifft(f)));
    neg = (f<=0);
    f(neg) = 0;

% Making sure the final joint density integrates to 1
    w = trapz(y,f,1);
    w = trapz(z,w);
    f = f/w;
    
    % Fixing numerical problems:
    p1 = (abs(Z) > 0.5);
    f(p1) = 0; % Very dangerous correction. Use with caution.
    
%     if (mod(j,5) == 0)
%     figure;
%     surf(y,z,f)
%     end
end

%% Calculating the risk neutral expectation, using the trapesoidal rule
V = trapz(z,f);
price = zeros(1,length(strikes));
for i = 1:length(strikes)
    func = max(y - strikes(i),0).*V;
    price(1,i) = exp(-r*T)*trapz(y,func);
end

end