function [price] = pi_fft_asian_vg(s0,strikes,sigma,lambda,vggamma,mu,r,q,T,dt,N)
% Note: Calling gamma for vggamma to avoid problems using the gamma
% function
%% Parameters
omega = lambda*log(1 - 1/2*sigma^2/lambda - vggamma/lambda);
mu = r - q + omega; % Mean correction.
m = round(T/dt); % Number of time steps
T = m*dt; % To avoid rounding error

%% Initialising grid
zmin = -0.8127; % DAX limits from NIG-code
zmax = 0.8077;
% Initialising grid
dz = (zmax-zmin)/(N-1);
z = (zmin:dz:zmax);
ymin = 8651;
ymax = 16651;
dy = (ymax-ymin)/(N-1);
y = ymin:dy:ymax;
% Defining the y,z grid
[Y,Z] = meshgrid(y,z);

%% Defining the transformation variable, z_1 = u(y_2|z_2), for f_{A_2|Z_2}
u = 3*Y/s0 - 1 - exp(Z); % u(y_2|z_2) = ln(tranz1)
correction = (u <= 0); % because e^z_1 > 0
u(correction) = 10^(-60);
u = log(u); % transz1 = u(y_2|z_2)
ind = ones(N);
ind(correction) = 0;
uder = 3./(3*Y-s0-s0*exp(Z));

%% Calculating the inital desity f_{A_2,Z_2}(y_2,z_2):
% f_{Z1}(z2-z1):
f_trans = 2*lambda^(lambda*dt)*exp((Z-u-mu*dt)*vggamma/sigma^2)/...
    (sqrt(2*pi)*sigma*gamma(lambda*dt))...
    .* besselk( lambda*dt-1/2 , 1/sigma^2*sqrt((Z-u-dt*mu).^2*...
    (2*lambda*sigma^2+vggamma^2)) ) ./ ( (Z-u-mu*dt).^2/...
    (2*lambda*sigma^2+vggamma^2)).^(1/4-1/2*lambda*dt).*ind;
% f_{Z_1}((u(y_2|z_2)):
f = 2*lambda^(lambda*dt)*exp((u-mu*dt)*vggamma/sigma^2)/(sqrt(2*pi)*...
    sigma*gamma(lambda*dt)).*...
    besselk( lambda*dt-1/2 , 1/sigma^2*sqrt((u-dt*mu).^2*...
    (2*lambda*sigma^2+vggamma^2)) ) ./...
    ( (u-mu*dt).^2/(2*lambda*sigma^2+vggamma^2)).^(1/4-1/2*lambda*dt).*ind;
% f_{A_2,Z_2}(y_2,z_2) (overwrites f_{Z_1} which is no longer needed)
f = f_trans.*f.*abs(uder);
neg = (f < 0); % Indecies to negative values
f(neg) = 0;
nan = isnan(f);
f(nan) = 1e-60;
% Scaling - Securing that the joint pdf integrates to 1
w = trapz(z,f,1);
w = trapz(y,w);
f = f/w;

%% Initialising the characteristic function of the transition prob.
freq = mod(1/2+(0:(N-1))/(N),1) - 1/2;
v = freq * (2*pi/dz);
v = v';
phi = conj( (1 - 1i*vggamma/lambda*v + sigma^2/...
    (2*lambda)*v.^2).^(-dt*lambda) .* exp(1i*v*mu*dt ));

% % Smoothing and Fixing numerical issues:
% realpart = real(phi);
% imagpart = imag(phi);
% px = (round(N*0.5*0.7)):(round(N*0.5*1.3)); % Problem area indeces
% py = imagpart(px);
% py = smooth(py,length(px),'rloess');
% imagpart(px) = py;
% neg = (real(realpart)<=0);
% realpart(neg) = 0;
% phi = realpart + 1i*imagpart;

phi_m = repmat(phi,1,length(phi));

%% Recursions
for j = 3:m
    % Interpolating, setting f_i-1
    YI = (j/(j-1))*Y - s0/(j-1)*exp(Z);
    f = interp2(Y,Z,f,YI,Z,'spline');
    corr = isnan(f);
    f(corr) = 0;
    
    % Calculating the convolution
    f = fft(f);
    f = (phi_m.*f);
    f = (real(ifft(f)));
    neg = (f<=0);
    f(neg) = 0;

% Making sure the final joint density integrates to 1
    q = trapz(y,f,1);
    w = trapz(z,q);
    f = f/w;
    
    % Fixing numerical problems:
    p1 = (abs(Z) > 0.5);
    f(p1) = 0; % Very dangerous correction. Use with caution
    
%     if (mod(j,25) == 0)
%     figure;
%     surf(y,z,f)
%     end
end
q = trapz(y,f,1);
w = trapz(z,q);
f = f/w;


%% Calculating the risk neutral expectation, using the trapesoidal rule
V = trapz(z,f);
price = zeros(1,length(strikes));
for i = 1:length(strikes)
    func = max(y - strikes(i),0).*V;
    price(1,i) = exp(-r*T)*trapz(y,func);
end

end